from PIL import Image
from io import BytesIO
import asyncio
from datetime import datetime, UTC

def NOW():
    return datetime.now(UTC).strftime('%Y%m%d%H%M%S')

async def create_gif(frames: list) -> bytes:
    pil_frames = [Image.open(BytesIO(image)).convert("RGB") for image in frames]
    frame_one = pil_frames[0]
    image_bytes = BytesIO()
    frame_one.save(image_bytes, format="GIF", append_images=pil_frames, save_all=True, optimize=False, duration=200, loop=0)

    return image_bytes.getbuffer()

async def write_gif(frames: list, name: str, timestamp: str = NOW()) -> bytes:
    try:
        image_bytes = await create_gif(frames)
        filename = f'{name}-{timestamp}.gif'
        with open(filename, 'wb') as f:
            f.write(image_bytes)
        print(f'wrote {filename}')
    except Exception as e:
        raise e

    return image_bytes
    


"""Console script for swpclib."""
import click
import sys
from . import swpclib
import asyncio

class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

''' 
A dictionary of URL groups, keys of this dict are assumed to be names of valid groups to generate animations from
values should be a list of tuples where the 0th element is the url of the swpc endpoint, and the 1st element is a 
filename excluding the file extension, which will always be .gif
'''
urls = {
    'solar_images': [
        ('/products/animations/suvi-primary-094.json', 'goes_094'),
        ('/products/animations/suvi-primary-195.json', 'goes_195'),
        ('/products/animations/suvi-primary-304.json', 'goes_304'),
        ('/products/animations/suvi-primary-map.json', 'suvi-map'),
        ('/products/animations/sdo-hmii.json', 'sdo-hmii')
    ],
    'graphs': [
        ('/products/animations/d-rap_f15_global.json', 'd-rap_f15'),
        ('/products/animations/wam-ipe/wfs_ionosphere_new.json', 'global_ionosphere')
    ],
    'cme': [
        ('/products/animations/lasco-c2.json', 'lasco-c2'),
        ('/products/animations/lasco-c3.json', 'lasco-c3'),
    ],

    'aurora': [
        ('/products/animations/ovation_north_24h.json', 'ovation_north'),
        ('/products/animations/ovation_south_24h.json', 'ovation_south'),
    ],
    'geomagnetic': [
        ('/products/animations/geoelectric/US-Canada-1D.json', 'geomagnetic_north_america'),
    ],
}

async def get_animations(urls):
    swpc = swpclib.Runner()
    await asyncio.gather(*[swpc.gen_gif(endpoint=url[0], name=url[1], write=True) for url in urls], return_exceptions=False)

@click.command()
@click.option('--group', help="Group of images to generate")
@click.option('--product', help="A swpc product endpoint that is a list of animation frames")
@click.option('--name', help="Name used for filename creation")
def animations(group=None, product=None, name=None):
    if group and group in urls.keys():
        url_group = urls.get(group, [])
        print(f"Generating {len(url_group)} animations for group: {group}")
        asyncio.run(get_animations(url_group))
    elif group:
            print(f"{group} is not a valid group. valid groups are: {[group for group in urls.keys()]}")
    elif product and name:
        print(f"Generating animation for: {name}")
        asyncio.run(get_animations([(product, name)]))


@click.command()
def space_weather():
    """Console script for swpclib."""
    runner = swpclib.Runner()
    data = asyncio.run(runner.get_standard())
    string = f"""\n{bcolors.OKBLUE}Current Space Weather:{bcolors.ENDC}

\tSFI: {data.get('sfi_data', {}).get('sfi')}
\tKP Index: {data.get('kp_index_data', {}).get('kp_index')}
\tA Index: {data['a_index_data']['a_index']}
\tSunspot Number: {data.get("ssn_data",{}).get("ssn")}
\tX-Class Probability: {data.get('probabilities_data',{})[0].get('x_class_1_day')}
    """
    print(string)
    alerts = [alert for alert in data["alerts"]]
    if "active" in data:
        alert_string = f"{bcolors.OKBLUE}Recent Alerts{bcolors.WARNING} (ACTIVE){bcolors.OKBLUE}:{bcolors.ENDC}\n\n"
    else:
        alert_string = f"{bcolors.OKBLUE}Recent Alerts:{bcolors.ENDC}\n\n"
    for alert in alerts:
        if "active" in alert:
            alert_string += (
                f"{bcolors.WARNING}{alert['formatted_message']}{bcolors.ENDC}"
            )
        else:
            alert_string += alert["formatted_message"]
        alert_string += "\n\n\n"
    alert_string += "Via NOAA Space Weather Prediction Center"
    print(alert_string)
    return 0


if __name__ == "__main__":
    sys.exit(space_weather())  # pragma: no cover



from datetime import datetime
import re

watch_sample = {
    "product_id": "A20F",
    "issue_datetime": "2022-09-27 19:57:05.470",
    "message": "Space Weather Message Code: WATA20\r\nSerial Number: 918\r\nIssue Time: 2022 Sep 27 1957 UTC\r\n\r\nWATCH: Geomagnetic Storm Category G1 Predicted\r\n\r\nHighest Storm Level Predicted by Day:\r\nSep 28:  None (Below G1)   Sep 29:  None (Below G1)   Sep 30:  G1 (Minor)\r\n\r\nTHIS SUPERSEDES ANY\\/ALL PRIOR WATCHES IN EFFECT\r\n\r\nNOAA Space Weather Scale descriptions can be found at\r\nwww.swpc.noaa.gov/noaa-scales-explanation\r\n\r\nPotential Impacts: Area of impact primarily poleward of 60 degrees Geomagnetic Latitude.\r\nInduced Currents - Weak power grid fluctuations can occur.\r\nSpacecraft - Minor impact on satellite operations possible.\r\nAurora - Aurora may be visible at high latitudes, i.e., northern tier of the U.S. such as northern Michigan and Maine.",
}

warn_sample = {
    "product_id": "K05W",
    "issue_datetime": "2022-09-27 09:34:10.330",
    "message": "Space Weather Message Code: WARK05\r\nSerial Number: 1682\r\nIssue Time: 2022 Sep 27 0934 UTC\r\n\r\nEXTENDED WARNING: Geomagnetic K-index of 5 expected\r\nExtension to Serial Number: 1681\r\nValid From: 2022 Sep 27 0122 UTC\r\nNow Valid Until: 2022 Sep 27 1800 UTC\r\nWarning Condition: Persistence\r\n\r\nNOAA Space Weather Scale descriptions can be found at\r\nwww.swpc.noaa.gov/noaa-scales-explanation\r\n\r\nPotential Impacts: Area of impact primarily poleward of 60 degrees Geomagnetic Latitude.\r\nInduced Currents - Weak power grid fluctuations can occur.\r\nSpacecraft - Minor impact on satellite operations possible.\r\nAurora - Aurora may be visible at high latitudes, i.e., northern tier of the U.S. such as northern Michigan and Maine.",
}

alert_sample = {
    "product_id": "K05A",
    "issue_datetime": "2022-09-27 01:32:19.593",
    "message": "Space Weather Message Code: ALTK05\r\nSerial Number: 1394\r\nIssue Time: 2022 Sep 27 0132 UTC\r\n\r\nALERT: Geomagnetic K-index of 5\r\nThreshold Reached: 2022 Sep 27 0131 UTC\r\nSynoptic Period: 0000-0300 UTC\r\n \r\nActive Warning: Yes\r\nNOAA Scale: G1 - Minor\r\n\r\nNOAA Space Weather Scale descriptions can be found at\r\nwww.swpc.noaa.gov/noaa-scales-explanation\r\n\r\nPotential Impacts: Area of impact primarily poleward of 60 degrees Geomagnetic Latitude.\r\nInduced Currents - Weak power grid fluctuations can occur.\r\nSpacecraft - Minor impact on satellite operations possible.\r\nAurora - Aurora may be visible at high latitudes, i.e., northern tier of the U.S. such as northern Michigan and Maine.",
}


def parse_message(output):

    _warning = re.compile(r".+W$")
    _alert = re.compile(r".+A$")
    _watch = re.compile(r".+F$")
    prid = output["product_id"]

    if _warning.match(prid):
        output["type"] = "warning"

    elif _watch.match(prid):
        output["type"] = "watch"

    elif _alert.match(prid):
        output["type"] = "alert"
    message = output["raw_message"]
    try:
        output["message_code"] = re.search(
            r"Space Weather Message Code: (.*)\r", message
        ).group(1)
    except AttributeError:
        pass
    try:
        output["serial_number"] = int(
            re.search(r"Serial Number: (.*)\r", message).group(1)
        )
    except AttributeError:
        pass
    try:
        _watch = re.compile(r"WATCH: (.*)\n.*\:", re.MULTILINE | re.DOTALL)
        output["watch"] = _watch.search(message).group(1)
    except AttributeError:
        pass
    try:
        _impacts = re.compile(
            r"Potential Impacts: (.*).*$", re.DOTALL | re.MULTILINE
        )
        output["impacts"] = _impacts.search(message).group(1)
    except AttributeError:
        pass
    try:
        _desc = re.compile(r"Description: (.*).*$", re.DOTALL | re.MULTILINE)
        output["description"] = _desc.search(message).group(1)
    except AttributeError:
        pass
    try:
        output["synoptic_period"] = re.search(
            r"\n.*Synoptic Period: (.*)\r", message
        ).group(1)
    except AttributeError:
        pass
    try:
        output["estimated_velocity"] = re.search(
            r"\n.*Estimated Velocity: (.*km)", message
        ).group(1)
    except AttributeError:
        pass
    try:
        output["threshold_reached"] = datetime.strptime(
            re.search(r"\nThreshold Reached: (.*)\r", message).group(1),
            "%Y %b %d %H%M %Z",
        )
    except AttributeError:
        pass
    try:
        _alert = re.compile(r"ALERT: (.*)\r")
        output["alert"] = _alert.search(message).group(1)
    except AttributeError:
        pass
    try:
        output["valid_from"] = datetime.strptime(
            re.search(r"Valid From: (.*)\r", message).group(1),
            "%Y %b %d %H%M %Z",
        )
        output["valid_until"] = datetime.strptime(
            re.search(r"\n.*Valid Until: (.*)\r", message).group(1),
            "%Y %b %d %H%M %Z",
        )
    except AttributeError:
        pass
    try:
        _watch = re.compile(r"\n.*WARNING: (.*)\r")
        output["warning"] = _watch.search(message).group(1)
    except AttributeError:
        pass
    try:
        if "valid_until" in output:
            if datetime.now() <= output["valid_until"]:
                output["active"] = True
    except AttributeError:
        pass
    alert_string = ""
    alert_string += f"{output['serial_number']} -- {output['message_code']} -- {output['issue_datetime']} -------\n\n"
    if "alert" in output:
        alert_string += output["alert"] + "\n"
    elif "warning" in output:
        alert_string += output["warning"] + "\n"
    elif "watch" in output:
        alert_string += output["watch"] + "\n"
    if "valid_from" in output:
        alert_string += (
            f"From: {output['valid_from']} --> {output.get('valid_until')}\n"
        )
    if "synoptic_period" in output:
        alert_string += f"Synoptic Period: {output['synoptic_period']}\n"
    if "threshold_reached" in output:
        alert_string += f"Threshold Reached: {output['threshold_reached']}\n"
    if "estimated_velocity" in output:
        alert_string += f"Estimated Velocity: {output['estimated_velocity']}\n"
    if "description" in output:
        alert_string += output["description"] + "\n"
    if "impacts" in output:
        alert_string += output["impacts"] + "\n"
    alert_string += "----------------------------------------------------"
    output["formatted_message"] = alert_string

    return output


def format_message(alert):
    """Return a formatted alert message"""

    output = {}
    output["product_id"] = alert["product_id"]
    output["issue_datetime"] = datetime.strptime(
        alert["issue_datetime"], "%Y-%m-%d %H:%M:%S.%f"
    )
    output["raw_message"] = alert["message"]
    output = parse_message(output)

    return output


if __name__ == "__main__":
    print(format_message(watch_sample))
    print("\n\n\n")
    print(format_message(warn_sample))
    print("\n\n\n")
    print(format_message(alert_sample))



import asyncio
import aiohttp
from . import alerts
from .animate import create_gif, write_gif

base_url = "https://services.swpc.noaa.gov"


class Runner:
    def __init__(self):
        self.base_url = base_url

    async def get_data_method(self, url):
        try:
            async with aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(total=20.0)
            ) as session:
                async with session.get(self.base_url + url) as response:
                    result = await response.json()
        except Exception as e:
            raise e

        return result

    async def get_bytes_method(self, url: str) -> bytes:
        try:
            async with asyncio.Semaphore(5):
                async with aiohttp.ClientSession() as session:
                    async with session.get(self.base_url + url) as response:
                        result = await response.content.read()
        except Exception as e:
            raise e

        return result

    async def fetch_all(self, urls: list, method) -> list:
        results = await asyncio.gather(*[method(url) for url in urls], return_exceptions=False)
        return results

    async def gen_gif(self, endpoint: str, name = None, write: bool = False) -> bytes: 
        resp = await self.get_data_method(endpoint)
        urls = [frame['url'] for frame in resp]
        frames = await self.fetch_all(urls, self.get_bytes_method)
        if write:
            gif = await write_gif(frames, name) 
        else:
            gif = await create_gif(frames) 

        return gif

    async def get_sfi(self, start=0, end=1, step=1):
        """get solar flux index"""

        try:
            data = await self.get_data_method("/json/f107_cm_flux.json")
            data_range = slice(start, end, step)
            sfi_data = {
                "sfi_data": {
                    "sfi": float(item["flux"] or float("-1")),
                    "timestamp": item["time_tag"],
                }
                for item in data[data_range]
            }
        except Exception as e:
            print(repr(e))
            sfi_data = {"sfi_data": None}

        return sfi_data

    async def get_ki(self, start=0, end=1, step=1):
        """get boulder K index"""

        try:
            data = await self.get_data_method("/json/boulder_k_index_1m.json")
            data_range = slice(start, end, step)
            ki_data = {
                "k_index_data": {
                    "k_index": float(item["k_index"]),
                    "timestamp": item["time_tag"],
                }
                for item in data[data_range]
            }
        except Exception as e:
            print(repr(e))
            ki_data = {"ki_data": None}

        return ki_data

    async def get_a(self, start=0, end=1, step=1):
        """get Fredricksburg A index"""

        try:
            data = await self.get_data_method(
                "/json/predicted_fredericksburg_a_index.json"
            )
            data_range = slice(start, end, step)
            a_data = {
                "a_index_data": {
                    "a_index": float(item["afred_1_day"]),
                    "a_2_day_index": float(item["afred_2_day"]),
                    "a_3_day_index": float(item["afred_3_day"]),
                    "timestamp": item["date"],
                }
                for item in data[data_range]
            }
        except Exception as e:
            print(repr(e))
            a_data = {"a_data": None}

        return a_data

    async def get_goes_195_angstroms_animation(self, write: bool = False):
        """Get 195 angstrom GOES image gif"""
        return await self.gen_gif(endpoint="/products/animations/suvi-primary-195.json", name="goes_195", write=write)

    async def get_goes_304_angstroms_animation(self, write: bool = False):
        """Get 304 angstrom GOES image gif"""
        return await self.gen_gif(endpoint="/products/animations/suvi-secondary-304.json", name="goes_304", write=write)

    async def get_goes_094_angstroms_animation(self, write: bool = False):
        """Get 094 angstrom GOES image gif"""
        return await self.gen_gif(endpoint="/products/animations/suvi-secondary-094.json", name="goes_094", write=write)

    async def get_ovation_north_24h_animation(self, write: bool = False):
        """Get Ovation Northern Hemisphere gif"""
        return await self.gen_gif(endpoint="/products/animations/ovation_north_24h.json", name="ovation_north", write=write)

    async def get_ovation_south_24h_animation(self, write: bool = False):
        """Get Ovation Southern Hemisphere gif"""
        return await self.gen_gif(endpoint="/products/animations/ovation_south_24h.json", name="ovation_south", write=write)

    async def get_ssn(self, start=0, end=1, step=1):
        """get observed sunspot number"""

        try:
            data = await self.get_data_method(
                "/json/solar-cycle/swpc_observed_ssn.json"
            )
            data_range = slice(start, end, step)
            data.reverse()
            ssn_data = {
                "ssn_data": {
                    "ssn": float(item["swpc_ssn"]),
                    "timestamp": item["Obsdate"],
                }
                for item in data[data_range]
            }
        except Exception as e:
            print(repr(e))
            ssn_data = {"ssn_data": None}

        return ssn_data

    async def get_kp(self, start=0, end=1, step=1):
        """get planetary k index"""

        try:
            data = await self.get_data_method("/json/planetary_k_index_1m.json")
            data.reverse()
            data_range = slice(start, end, step)
            kp_data = {
                "kp_index_data": {
                    "kp_index": float(item["kp_index"]),
                    "timestamp": item["time_tag"],
                }
                for item in data[data_range]
            }
        except Exception as e:
            print(repr(e))
            kp_data = {"kp_data": None}

        return kp_data

    async def get_probabilities(self, start=0, end=1, step=1):
        """get solar event probabilities"""

        try:
            data = await self.get_data_method("/json/solar_probabilities.json")
            data_range = slice(start, end, step)
            probabilities_data = {
                "probabilities_data": [item for item in data[data_range]]
            }
        except Exception as e:
            print(repr(e))
            probabilities_data = {"probabilities_data": None}

        return probabilities_data

    async def get_alerts(self, start=0, end=3, step=1):
        """Returns solar weather alerts"""

        try:
            data = await self.get_data_method("/products/alerts.json")
            data_range = slice(start, end, step)
            alerts_data = []
            active = False
            for alert in data[data_range]:
                try:
                    a = alerts.format_message(alert)
                    if "active" in a:
                        active = True
                    alerts_data.append(a)
                except Exception as e:
                    print(repr(e))
            alerts_data = {"alerts": alerts_data}
            if active:
                alerts_data["active"] = True
        except Exception as e:
            print(repr(e))
            alerts_data = {"alerts": None}

        return alerts_data

    async def get_standard(self, start=0, end=1, step=1):
        """return standard output"""

        try:
            standard_group = await asyncio.gather(
                self.get_sfi(start, end, step),
                self.get_kp(start, end, step),
                self.get_probabilities(start, end, step),
                self.get_ssn(start, end, step),
                self.get_a(start, end, step),
                self.get_alerts(),
            )
            data = {}
            for _dict in standard_group:
                data.update(_dict)
        except Exception as e:
            print(repr(e))
            data = None
        return data


if __name__ == "__main__":
    swpc = Runner()
    print(asyncio.run(swpc.get_standard()))
